from ..rainbows import *
from .setup_tests import *


def test_rainbow_basics():

    nw, nt = 24, 72
    r = Rainbow(
        wavelength=np.linspace(1, 2, nw) * u.micron,
        time=np.linspace(-1, 1, nt) * u.hour,
        flux=np.ones((nw, nt)),
    )

    # test the basic sizes and shapes
    assert r.shape == (nw, nt)
    assert r.nwave == nw
    assert r.ntime == nt
    assert r.nflux == nw * nt

    # testing slicing, indexing, masking
    assert r[:, :].shape == (nw, nt)
    assert r[::2, ::3].shape == (nw / 2, nt / 3)
    assert r[2:4, :].shape == (2, nt)
    assert r[np.arange(2, 4), :].shape == (2, nt)
    assert r[r.wavelength < 1.5 * u.micron, :].shape == (nw / 2, nt)
    assert r[:, r.time < 0 * u.day].shape == (nw, nt / 2)
    assert r[0, :].shape == (1, nt)
    assert r[:, 0].shape == (nw, 1)
    assert r[0, 0].shape == (1, 1)

    assert r[r.wavelength < 1.5 * u.micron, r.time < 0 * u.day].shape == (
        nw / 2,
        nt / 2,
    )


def test_essential_properties():
    # create a simulated rainbow
    r = SimulatedRainbow().inject_noise()

    # wavelength
    x = r.wavelength
    r.wavelength = x * 2
    assert np.all(r.wavelength == x * 2)

    # time
    x = r.time
    r.time = x * 2
    assert np.all(r.time == x * 2)

    # flux
    x = r.flux
    r.flux = x * 2
    assert np.all(r.flux == x * 2)

    # uncertainty
    x = r.uncertainty
    r.uncertainty = x * 2
    assert np.all(r.uncertainty == x * 2)

    # ok
    r.ok = np.isfinite(r.flux) == False
    assert np.all(r.ok == False)


def test_automatic_quantity_sorting():
    # create a simulated rainbow
    r = SimulatedRainbow().inject_noise()

    # does a wavelength-shaped array end up in the right spot?
    r.background = np.random.normal(10, 1, r.nwave)
    assert "background" in r.wavelike

    # does a time-shaped array end up in the right spot?
    r.temperature = np.random.normal(10, 1, r.ntime)
    assert "temperature" in r.timelike

    # does a fluxes-shaped array end up in the right spot?
    r.saturated = np.random.normal(0, 1, r.shape) > 2
    assert "saturated" in r.fluxlike

    # does a weirdly-shaped array still get assigned somewhere?
    r.weird = np.random.normal(0, 1, (1, 2, 3, 4))
    r.weird + 1


def test_shape_warnings():
    a = SimulatedRainbow().inject_noise()
    with pytest.warns(match="transpose"):
        a.flux = a.flux.T

    b = SimulatedRainbow().inject_noise()
    with pytest.warns(match="shape"):
        b.wavelength = np.arange(b.nwave + 1)

    c = SimulatedRainbow().inject_noise()
    with pytest.warns(match="shape"):
        c.time = np.arange(c.ntime + 1)

    d = SimulatedRainbow().inject_noise()
    with pytest.warns(match="shape"):
        d.uncertainty = np.ones((d.nwave + 1, d.ntime + 1))

    e = SimulatedRainbow().inject_noise()
    with pytest.warns(match="shape"):
        d.ok = np.ones((d.nwave + 1, d.ntime + 1))


def test_sort():
    w = np.linspace(2, 1) * u.micron
    t = np.linspace(1, -1) * u.day

    with pytest.warns(match="input times were not monotonically increasing"):
        r = SimulatedRainbow(time=t).inject_noise()
        r._validate_core_dictionaries()
        print(r.wavelength)
        print(r.time)
        print(r.original_time_index)
        assert np.all(r.time == t[::-1])

    with pytest.warns(match="input wavelengths were not monotonically increasing"):
        r = SimulatedRainbow(wavelength=w).inject_noise()
        r._validate_core_dictionaries()
        print(r.wavelength)
        print(r.time)
        print(r.original_wave_index)
        assert np.all(r.wavelength == w[::-1])


def test_help():
    Rainbow().help()



from ..imports import *
from ..resampling import *
from ..resampling import leftright_to_edges
from .setup_tests import *


def test_resampling():

    resample_directory = os.path.join(test_directory, "core-resample-examples")
    try:
        os.mkdir(resample_directory)
    except FileExistsError:
        pass

    N = 11
    x_edges = np.linspace(2, 4, N)
    x = (x_edges[1:] + x_edges[:-1]) / 2
    y = np.random.normal(x, 0.5)
    N_new = 29
    x_new_edges = np.linspace(1.6, 4.3, N_new)
    x_new = (x_new_edges[1:] + x_new_edges[:-1]) / 2

    a = resample_while_conserving_flux(xin=x, yin=y, xout=x_new, visualize=True)
    plt.savefig(
        os.path.join(
            resample_directory, "demonstration-of-resampling-centers-in-centers-out.pdf"
        )
    )

    b = resample_while_conserving_flux(
        yin=y, xout=x_new, xin_edges=x_edges, visualize=True
    )
    plt.savefig(
        os.path.join(
            resample_directory, "demonstration-of-resampling-edges-in-centers-out.pdf"
        )
    )

    c = resample_while_conserving_flux(
        yin=y, xin_edges=x_edges, xout_edges=x_new_edges, visualize=True
    )
    plt.savefig(
        os.path.join(
            resample_directory, "demonstration-of-resampling-edges-in-edges-out.pdf"
        )
    )

    d = resample_while_conserving_flux(
        yin=y, xin=x, xout_edges=x_new_edges, visualize=True
    )
    plt.savefig(
        os.path.join(
            resample_directory, "demonstration-of-resampling-centers-in-edges-out.pdf"
        )
    )

    for x in [b, c, d]:
        assert np.isclose(np.sum(a["y"]), np.sum(x["y"]))
        assert np.all(np.isclose(a["x_edge_lower"], x["x_edge_lower"]))
        assert np.all(np.isclose(a["x_edge_upper"], x["x_edge_upper"]))


def test_bintogrid(N_original=23):

    bin_directory = os.path.join(test_directory, "binning-examples")
    try:
        os.mkdir(bin_directory)
    except FileExistsError:
        pass

    def save_binning_example_figure():
        f = plt.gca().get_title()
        f = "demonstration-of-" + f.replace(" ", "") + ".pdf"
        f = f.replace("$\sigma$", "uncertainty")
        f = f.replace(",", "-")
        plt.savefig(os.path.join(bin_directory, f))

    for input_grid in ["uniform", "irregular"]:
        if input_grid == "uniform":
            x = np.linspace(-3, 3, N_original)
        elif input_grid == "irregular":
            x = np.sort(np.random.uniform(-3, 3, N_original))

        for uncertainties in ["no", "with"]:
            plt.close("all")
            if uncertainties == "no":
                u = None
                y = np.exp(-0.5 * x**2)
            else:
                u = np.ones(N_original) * 0.2
                y = np.random.normal(np.exp(-0.5 * x**2), u)

            for how in ["bigger", "smaller"]:
                if how == "bigger":
                    dx = np.median(np.diff(x)) * np.random.uniform(3, 5)
                elif how == "smaller":
                    dx = np.median(np.diff(x)) / np.random.uniform(3, 5)

                label = f"{input_grid} input grid, {uncertainties} $\sigma$, new bins are {how}"

                # do different options give same answer?
                a = bintogrid(x, y, unc=u, dx=dx, visualize=True)
                plt.title(f"bin to dx={dx:.3}, {label}")
                save_binning_example_figure()

                centers = a["x"]
                b = bintogrid(x, y, unc=u, newx=centers, visualize=True)
                plt.title(f"bin to {len(centers)} newx, {label}")
                save_binning_example_figure()
                assert np.all(np.isclose(a["y"], b["y"]))

                edges = leftright_to_edges(a["x_edge_lower"], a["x_edge_upper"])
                c = bintogrid(x, y, unc=u, newx_edges=edges, visualize=True)
                plt.title(f"bin to {len(edges)} newx_edges, {label}")
                save_binning_example_figure()
                assert np.all(np.isclose(a["y"], c["y"]))

            label = f"{input_grid} input grid, {uncertainties} " + r"$\sigma$"

            # does binning to self give self?
            d = bintogrid(x, y, unc=u, newx=x, visualize=True)
            plt.title(f"bin to self, {label}")
            save_binning_example_figure()
            assert np.all(np.isclose(d["x"], x))
            assert np.all(np.isclose(d["y"], y))

            # does binning by "N points" work, and match other modes?
            for nx in [1, 2, 3, 5]:
                e = bintogrid(x, y, unc=u, nx=nx, visualize=True)
                plt.title(f"bin by {nx} bins, {label}")
                save_binning_example_figure()
    plt.close("all")



from .setup_tests import *
from ..imports import *
from ..spectra import *


def test_spectral_library_R(cmap=one2another("indigo", "tomato"), N=5):
    fi, ax = plt.subplots(
        3, 1, figsize=(8, 6), dpi=300, sharex=True, sharey=True, constrained_layout=True
    )
    for R, a in zip([10, 100, 1000], ax):
        plt.sca(a)
        for i, T in enumerate(
            np.round(np.logspace(np.log10(2300), np.log10(12000), N))
        ):
            color = cmap(i / (N - 1))
            w, f = get_phoenix_photons(temperature=T, R=R)
            plt.loglog(w, f, color=color)
            w, f = get_planck_photons(temperature=T, R=R)
            plt.plot(w, f, color=color, linestyle=":")
            plt.ylim(1e21, 1e25)
            plt.xlim(0.05, 5)
            plt.text(0.98, 0.92, f"R={R}", transform=a.transAxes, ha="right", va="top")
    fi.supxlabel(f"Wavelength ({w.unit.to_string('latex_inline')})")
    fi.supylabel(f"Surface Flux ({f.unit.to_string('latex_inline')})")
    plt.savefig(
        os.path.join(
            test_directory, "demonstration-of-spectral-library-with-constant-R.pdf"
        )
    )


def test_spectral_library_wavelengths(cmap=one2another("indigo", "tomato"), N=5):
    fi, ax = plt.subplots(
        3, 1, figsize=(8, 6), dpi=300, sharex=True, sharey=True, constrained_layout=True
    )
    for how_many, a in zip([10, 100, 1000], ax):
        plt.sca(a)
        wavelength = np.logspace(-1, 0, how_many) * u.micron
        for i, T in enumerate(
            np.round(np.logspace(np.log10(2300), np.log10(12000), N))
        ):
            color = cmap(i / (N - 1))
            w, f = get_phoenix_photons(temperature=T, wavelength=wavelength)
            plt.loglog(w, f, color=color)
            w, f = get_planck_photons(temperature=T, wavelength=wavelength)
            plt.plot(w, f, color=color, linestyle=":")
            plt.ylim(1e21, 1e25)
            plt.xlim(0.05, 2)
            plt.text(
                0.98, 0.92, f"N={how_many}", transform=a.transAxes, ha="right", va="top"
            )
    fi.supxlabel(f"Wavelength ({w.unit.to_string('latex_inline')})")
    fi.supylabel(f"Surface Flux ({f.unit.to_string('latex_inline')})")
    plt.savefig(
        os.path.join(
            test_directory,
            "demonstration-of-spectral-library-with-custom-wavelengths.pdf",
        )
    )


def test_spectral_library_loads_correct_R():
    for i in range(2):
        w = np.linspace(0.7, 0.71, 1000) * u.micron
        wamb, s_amb = get_phoenix_photons(
            temperature=int(3900.0), wavelength=w, logg=4.4, metallicity=0.0
        )
        wspot, s_spot = get_phoenix_photons(
            temperature=int(3100.0), wavelength=w, logg=4.4, metallicity=0.0
        )
        plt.figure()
        plt.plot(wspot, s_spot)
        plt.plot(wamb, s_amb)
    plt.savefig(
        os.path.join(
            test_directory,
            "demonstration-of-spectral-library-loading-minimum-necessary-resolution.pdf",
        )
    )



from ..rainbows import *
from .setup_tests import *


def test_rainbow_npy():
    filename = os.path.join(test_directory, "test.rainbow.npy")
    a = SimulatedRainbow().inject_noise()
    a.save(filename)
    b = Rainbow(filename)
    assert a == b


def test_rainbow_FITS():
    filename = os.path.join(test_directory, "test.rainbow.fits")
    a = SimulatedRainbow().inject_noise()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        a.save(filename, overwrite=True)
    b = read_rainbow(filename)
    assert a == b


def test_text():
    filename = os.path.join(test_directory, "test.rainbow.txt")
    a = SimulatedRainbow().inject_noise()
    a.save(filename, overwrite=True)
    b = read_rainbow(filename)
    assert a == b


def test_xarray():

    for f in [
        "stellar-spec-planettest-modetest-codechromatic-authorzkbt.xc",
        "raw-light-curves-planettest-modetest-codechromatic-authorzkbt.xc",
        "fitted-light-curves-planettest-modetest-codechromatic-authorzkbt.xc",
    ]:
        filename = os.path.join(test_directory, f)
        print(filename)
        a = SimulatedRainbow().inject_transit().inject_systematics().inject_noise()

        with pytest.warns(match="required metadata keyword"):
            a.save(filename)
        b = read_rainbow(filename)
        assert a == b


def test_rainbow_npy():
    filename = os.path.join(test_directory, "test.rainbow.npy")
    a = SimulatedRainbow().inject_noise()
    a.save(filename)
    b = Rainbow(filename)
    assert a == b


def test_guess_readers():

    assert guess_reader("some-neat-file.rainbow.npy") == from_rainbow_npy
    assert guess_reader("some-neat-file-x1dints.fits") == from_x1dints

    # eureka readers
    assert guess_reader("S3_wasp39b_ap6_bg7_SpecData.h5") == from_eureka_S3
    assert guess_reader("S4_wasp39b_ap6_bg7_LCData.h5") == from_eureka_S4
    assert (
        guess_reader(
            [
                "S5_wasp39b_ap6_bg7_Table_Save_ch0.txt",
                "S5_wasp39b_ap6_bg7_Table_Save_ch1.txt",
            ]
        )
        == from_eureka_S5
    )

    # xarray common format
    assert guess_reader("stellar-spec-wow.xc") == from_xarray_stellar_spectra
    assert guess_reader("raw-light-curves-wow.xc") == from_xarray_raw_light_curves
    assert guess_reader("fitted-light-curve-wow.xc") == from_xarray_fitted_light_curves


def test_guess_writers():

    assert guess_writer("some-neat-file.rainbow.npy") == to_rainbow_npy

    # xarray common format
    assert guess_writer("stellar-spec-wow.xc") == to_xarray_stellar_spectra
    assert guess_writer("raw-light-curves-wow.xc") == to_xarray_raw_light_curves
    assert guess_writer("fitted-light-curve-wow.xc") == to_xarray_fitted_light_curves



from ..imports import *

__all__ = ["calculate_planck_flux", "get_planck_photons"]


def calculate_planck_flux(wavelength, temperature):
    """
    Calculate the surface flux from a thermally emitted surface,
    according to Planck function.

    Parameters
    ----------
    wavelength : u.Quantity
        The wavelengths at which to calculate,
        with units of wavelength.
    temperature : u.Quantity
        The temperature of the thermal emitter,
        with units of K.

    Returns
    -------
    surface_flux : u.Quantity
        The surface flux, evaluated at the wavelengths.
    """

    # define variables as shortcut to the constants we need
    h = con.h
    k = con.k_B
    c = con.c

    # the thing that goes in the exponent (its units better cancel!)
    z = h * c / (wavelength * k * temperature)

    # calculate the intensity from the Planck function
    intensity = (2 * h * c**2 / wavelength**5 / (np.exp(z) - 1)) / u.steradian

    # calculate the flux assuming isotropic emission
    flux = np.pi * u.steradian * intensity

    # return the intensity
    return flux.to("W/(m**2*nm)")


def get_planck_photons(
    temperature=3000, wavelength=None, R=100, wlim=[0.04, 6] * u.micron, **kw
):
    """
    Calculate the surface flux from a thermally emitted surface,
    according to Planck function, in units of photons/(s * m**2 * nm).

    (Note, this evaluates the Planck function at the exact
    wavelength values; it doesn't do anything fancy to integrate
    over binwidths, so if you're using very wide (R~a few) bins
    your integrated fluxes will be messed up.)

    Parameters
    ----------
    temperature : u.Quantity
        The temperature of the thermal emitter,
        with units of K.
    wavelength : u.Quantity
        The wavelengths at which to calculate,
        with units of wavelength.
    R : float
        The spectroscopic resolution for creating a log-uniform
        grid that spans the limits set by `wlim`, only if
        `wavelength` is not defined.
    wlim : u.Quantity
        The two-element [lower, upper] limits of a wavelength
        grid that would be populated with resolution `R`, only if
        `wavelength` is not defined.
    kw : dict
        Other keyword arguments will be ignored.

    Returns
    -------
    photons : u.Quantity
        The surface flux in units of photons
    """

    # make sure the temperature unit is good (whether or not it's supplied)
    temperature_unit = u.Quantity(temperature).unit
    if temperature_unit == u.K:
        temperature_with_unit = temperature
    elif temperature_unit == u.Unit(""):
        temperature_with_unit = temperature * u.K

    # create a wavelength grid if one isn't supplied
    if wavelength is None:
        wavelength_unit = wlim.unit
        wavelength = (
            np.exp(np.arange(np.log(wlim[0].value), np.log(wlim[1].value), 1 / R))
            * wavelength_unit
        )

    energy = calculate_planck_flux(
        wavelength=wavelength, temperature=temperature_with_unit
    )
    photon_energy = con.h * con.c / wavelength / u.ph

    return wavelength, (energy / photon_energy).to(u.ph / (u.s * u.m**2 * u.nm))



from ....imports import *
from ..utilities import _add_panel_labels

__all__ = ["imshow_with_models"]


def _create_title(s):
    """
    A helper function to make better titles.
    """
    if s == "flux":
        return "Data"
    else:
        return s.replace("_", " ").title()


def imshow_with_models(
    self,
    models=["systematics_model", "planet_model"],
    vlimits_data=[0.98, 1.02],
    vspan_residuals=0.02,
    figsize=(8, 3),
    label="inside",
    label_textkw={},
    filename=None,
    **kw,
):
    """
    Produce a multi-panel imshow figure comparing
    the data to model components. The left panel
    will be the data ('flux'), the right panel will
    be the residuals ('flux' - 'model'), and the middle
    panel(s) will be the model(s) connecting them.

    Parameters
    ----------
    models : list
        The list of model quantities to display,
        each in its own panel. A popular option might be:
        `models = ['systematics_model', 'planet_model']
        Whatever you include here does not affect the
        values of the Residuals panel, which are always
        exactly 'flux' - 'model'.
    vlimits_data : list
        The value limits for the data and model colormap,
        specified as `vlimits_data = [vmin, vmax]`.
        If `None` or `[None, None]`, the limits will be
        replaced with the [1,99] percentiles of the flux.
    vspan_residuals : float
        The distance away from zero where the colormap
        for the residuals will cut off. If `None`, the
        span will be set to 3 standard deviations.
    figsize : tuple
        The figure size for the multipanel plot.
        It should probably be wider than it is tall.
    label : bool
        Should we add (a), (b), (c) labels to the panels,
        and where? `False` or `None` give no labels, 'inside'
        sets them in the upper left inside corner, 'outside'
        sets them above the upper left corner.
    label_textkw : dict
        Keywords to pass along to the `plt.text` command
        that makes the labels.
    **kw : dict
        All additional keywords will be passed along
        to `.imshow` for all panels. Please see the docstring
        for that function for the full set of options.
    """

    # make sure we use only models that exist
    models_that_exist = []
    for m in models:
        if self.get(m) is None:
            cheerfully_suggest(f"'{m}' doesn't exist and will be skipped.")
        else:
            models_that_exist.append(m)
    if len(models_that_exist) == 0:
        models_that_exist = ["model"]
    models = models_that_exist

    # make sure color limits are set for data and model
    percentiles = [1, 99]
    if vlimits_data is None:
        vlimits_data = np.percentile(self.flux, percentiles)
    elif vlimits_data[0] is None:
        vlimits_data[0] = np.percentile(self.flux, percentiles[0])
    elif vlimits_data[1] is None:
        vlimits_data[1] = np.percentile(self.flux, percentiles[1])

    # make sure color limits are set for the residuals
    if vspan_residuals is None:
        vspan_residuals = np.nanstd(self.residuals) * 3

    # set up the panels
    row_data, row_colorbar = 1, 0
    N = len(models) + 2
    fi, ax = plt.subplots(
        2,
        N,
        figsize=figsize,
        dpi=600,
        gridspec_kw=dict(height_ratios=[1, 20]),
        sharex="row",
        sharey="row",
        constrained_layout=True,
    )

    # decide whether to use imshow or pcolormesh
    if (self.wscale in ["linear", "log"]) and (self.tscale in ["linear", "log"]):
        show = self.imshow
    else:
        show = self.pcolormesh
    # plot the first three panels
    for i, k in enumerate(["flux"] + models):
        show(
            ax=ax[row_data, i],
            quantity=k,
            vmin=vlimits_data[0],
            vmax=vlimits_data[1],
            colorbar=False,
            **kw,
        )
        ax[row_data, i].set_title(_create_title(k))

    # set up the colorbar to span the data and models
    gs = ax[row_colorbar, 0].get_gridspec()
    for a in ax[row_colorbar, :-1]:
        a.remove()
    ax_colorbar = fi.add_subplot(gs[row_colorbar, :-1])
    plt.colorbar(cax=ax_colorbar, orientation="horizontal")

    # plot the residuals
    k = "residuals"
    show(
        ax[row_data, -1],
        k,
        vmin=-vspan_residuals,
        vmax=vspan_residuals,
        colorbar=False,
        **kw,
    )
    ax[row_data, -1].set_title("Residuals")

    # set up the colorbar for the residuals
    ax_colorbar_residuals = plt.subplot(gs[row_colorbar, -1])
    cbar = plt.colorbar(cax=ax_colorbar_residuals, orientation="horizontal")
    for a in [ax_colorbar, ax_colorbar_residuals]:
        a.tick_params(labelsize=8)
    x = vspan_residuals / 2
    cbar.set_ticks([-x, 0, x], labels=[f"{-x:.2}", "0", f"{x:.2}"])

    # clear all the axis labels and replace with shared
    for a in ax[row_data, :]:
        xlabel, ylabel = a.get_xlabel(), a.get_ylabel()
        a.set_ylabel("")
        a.set_xlabel("")
    fi.supxlabel(xlabel)
    fi.supylabel(ylabel)

    if label:
        _add_panel_labels(ax[row_data, :], preset=label, **label_textkw)
    plt.sca(ax[row_data, 0])
    if filename is not None:
        self.savefig(filename)



from ..rainbows import *
from .setup_tests import *


def test_trim():
    r = SimulatedRainbow().inject_noise()
    r.fluxlike["flux"][:3, :] = np.nan
    r.fluxlike["flux"][:, -4:] = np.nan
    original_shape = r.shape
    t = r.trim()
    new_shape = t.shape
    assert new_shape[0] == original_shape[0] - 3
    assert new_shape[1] == original_shape[1] - 4
    print(r, t)



from ..rainbows import *
from .setup_tests import *
from ..rainbows.visualizations import _add_panel_labels


def test_imshow():
    plt.close("all")

    plt.figure()
    SimulatedRainbow(R=20).imshow()
    plt.savefig(os.path.join(test_directory, "demonstration-of-imshow.pdf"))

    plt.figure()
    SimulatedRainbow(dw=0.2 * u.micron).imshow()

    plt.figure()
    Rainbow(
        wavelength=np.arange(1, 5) * u.micron,
        time=np.arange(1, 6) * u.hour,
        flux=np.ones((4, 5)),
    ).imshow()

    fi, ax = plt.subplots(2, 1, sharex=True)
    SimulatedRainbow(R=10).inject_noise().imshow(w_unit="nm", ax=ax[0])
    SimulatedRainbow(dw=0.2 * u.micron).inject_noise().imshow(ax=ax[1], w_unit="nm")
    plt.savefig(os.path.join(test_directory, "demonstration-of-imshow-units.pdf"))


def test_imshow_quantities():
    plt.close("all")

    s = SimulatedRainbow().inject_transit().inject_noise(signal_to_noise=500)
    for k in "abcde":
        s.fluxlike[k] = np.random.uniform(4, 5, s.shape)
    s.imshow_quantities(maxcol=1, panel_size=(8, 2))
    plt.savefig(os.path.join(test_directory, "demonstration-of-imshow-quantities.pdf"))


def test_plot():
    plt.close("all")

    SimulatedRainbow(R=10).inject_noise().plot()
    plt.savefig(os.path.join(test_directory, "demonstration-of-plot.pdf"))


def test_plot_unnormalized():
    plt.close("all")

    w = np.logspace(0, 1, 5) * u.micron
    plt.figure()
    s = SimulatedRainbow(wavelength=w, star_flux=w.value**2).inject_noise(
        signal_to_noise=5
    )
    s.plot(spacing=0)
    plt.savefig(
        os.path.join(
            test_directory,
            "demonstration-of-plot-without-normalization.pdf",
        )
    )


def test_plot_quantities():
    plt.close("all")

    r = SimulatedRainbow(R=10).inject_noise()
    for k in "abcdefg":
        r.timelike[f'timelike quantity "{k}"'] = np.random.normal(0, 1, r.ntime) * u.m
        r.wavelike[f'wavelike quantity "{k}"'] = np.random.normal(0, 1, r.nwave) * u.s

    for k in ["time", "wavelength"]:
        for x in [k, "index"]:
            r.plot_quantities(xaxis=k, what_is_x=x)
            plt.savefig(
                os.path.join(
                    test_directory,
                    f"demonstration-of-plot-quantities-data={k}-xaxis={x}.pdf",
                )
            )


def test_animate():
    plt.close("all")

    # test a transit, since along both dimensions
    d = SimulatedRainbow(dw=0.1 * u.micron, dt=5 * u.minute).inject_noise(
        signal_to_noise=1000
    )
    theta = np.linspace(0, 2 * np.pi, d.nwave)
    planet_radius = np.sin(theta) * 0.05 + 0.15
    e = d.inject_transit(planet_radius=planet_radius)
    scatterkw = dict()
    e.animate_lightcurves(
        filename=os.path.join(
            test_directory, "demonstration-of-animate-lightcurves.gif"
        ),
        scatterkw=scatterkw,
    )
    e.animate_spectra(
        filename=os.path.join(
            test_directory, "demonstration-of-animate-spectra-demonstration.gif"
        ),
        scatterkw=scatterkw,
    )


def test_animate_other_quantities():
    plt.close("all")

    k = "some-imaginary-fluxlike-quantity"
    s = (
        SimulatedRainbow(R=5, dt=20 * u.minute)
        .inject_transit()
        .inject_systematics(amplitude=0.001, fluxlike=[k])
        .inject_noise(signal_to_noise=1000)
    )
    s.animate_spectra(
        os.path.join(
            test_directory,
            "demonstration-of-animate-quantities-beside-flux-spectra.gif",
        ),
        quantity=k,
    )
    s.animate_lightcurves(
        os.path.join(
            test_directory,
            "demonstration-of--animate-quantities-beside-flux-lightcurves.gif",
        ),
        quantity=k,
    )


def test_cmap():
    plt.close("all")

    r = SimulatedRainbow(R=10).inject_noise()

    # can we set up the wavelength-based color map
    r.setup_wavelength_colors(cmap=one2another("black", "red"))

    # test a few examples
    assert r.get_wavelength_color(r.wavelength[0]) == (0.0, 0.0, 0.0, 1.0)
    assert r.get_wavelength_color(r.wavelength[-1]) == (1.0, 0.0, 0.0, 1.0)


def test_imshow_interact():
    plt.close("all")

    plt.figure()
    SimulatedRainbow(R=10).inject_noise().imshow_interact()


def test_plot_one_wavelength():
    plt.close("all")

    s = SimulatedRainbow(wavelength=[1] * u.micron).inject_noise()
    s.plot()


def test_imshow_one_wavelength():
    plt.close("all")

    with pytest.warns(match="hard to imshow "):

        s = SimulatedRainbow(wavelength=[1] * u.micron).inject_noise()
        ax = s.imshow()
        assert "Wavelength Index" in ax.get_ylabel()

        s = SimulatedRainbow().inject_noise()
        b = s.bin(nwavelengths=s.nwave)
        ax = b.imshow()
        assert "Wavelength (" in ax.get_ylabel()
        ylim = ax.get_ylim()
    plt.savefig(
        os.path.join(
            test_directory,
            f"demonstration-of-imshow-with-just-one-wavelength.pdf",
        )
    )


def test_imshow_randomized_axes():
    plt.close("all")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        s = (
            SimulatedRainbow(
                time=np.random.uniform(-3, 3, 12) * u.hour,
                wavelength=np.random.uniform(0.5, 5, 8) * u.micron,
            )
            .inject_transit()
            .inject_noise(signal_to_noise=300)
        )
        s.fluxlike["flux"] += 0.003 * s.wavelength.value[:, np.newaxis]

        fi, ax = plt.subplots(1, 3, figsize=(10, 3), constrained_layout=True)
        kw = dict(vmin=0.98, vmax=1.02)
        s.imshow(ax=ax[0], **kw)
        s.get_average_spectrum_as_rainbow().imshow(ax=ax[1], **kw)
        s.get_average_lightcurve_as_rainbow().imshow(ax=ax[2], **kw)
        for i in [0, 2]:
            assert "Time Index" in ax[i].get_xlabel()
        for i in [0, 1]:
            assert "Wavelength Index" in ax[1].get_ylabel()
        plt.savefig(
            os.path.join(
                test_directory,
                f"demonstration-of-imshow-with-irregular-axes.pdf",
            )
        )


def test_imshow_both_orientations():
    plt.close("all")

    s = (
        SimulatedRainbow(R=5, dt=10 * u.minute)
        .inject_transit(planet_radius=0.2)
        .inject_systematics(amplitude=0.005)
        .inject_noise(signal_to_noise=1000)
    )

    fi, ax = plt.subplots(2, 2, constrained_layout=True, figsize=(6, 6))
    for i, k in enumerate(["time", "wavelength"]):
        s.plot(ax=ax[0, i], xaxis=k)
        s.imshow(ax=ax[1, i], xaxis=k, colorbar=False)
        plt.savefig(
            os.path.join(
                test_directory,
                f"demonstration-of-imshow-with-both-orientations.pdf",
            )
        )


def test_pcolormesh_both_orientations():
    plt.close("all")

    s = (
        SimulatedRainbow(R=5, dt=10 * u.minute)
        .inject_transit(planet_radius=0.2)
        .inject_systematics(amplitude=0.005)
        .inject_noise(signal_to_noise=1000)
    )

    fi, ax = plt.subplots(2, 2, constrained_layout=True, figsize=(6, 6))
    for i, k in enumerate(["time", "wavelength"]):
        s.plot(ax=ax[0, i], xaxis=k)
        s.pcolormesh(ax=ax[1, i], xaxis=k, colorbar=False)
    plt.savefig(
        os.path.join(
            test_directory, "demonstration-of-pcolormesh-both-orientations.pdf"
        )
    )


def test_both_types_of_plot():
    plt.close("all")

    N, M = 10, 20
    r = (
        SimulatedRainbow(
            wavelength=np.linspace(1, 2, N) * u.micron,
            time=np.linspace(-3, 3, M) * u.hour,
            star_flux=np.linspace(1, 1.1, N),
        )
        .inject_transit(planet_radius=0.2)
        .inject_noise(signal_to_noise=500)
    )
    fi, ax = plt.subplots(4, 2, figsize=(8, 15), constrained_layout=True)
    for i, f in enumerate(["plot_lightcurves", "plot_spectra"]):
        getattr(r, f)(ax=ax[0, i])
        getattr(r.normalize(), f)(ax=ax[1, i], errorbar=True, spacing=0)
        getattr(r.normalize(), f)(ax=ax[2, i], spacing=0.01)
        getattr(r.normalize(), f)(
            ax=ax[3, i],
            spacing=0,
            errorbar=True,
            plotkw=dict(color="orchid"),
            textkw=dict(color="orchid"),
            errorbarkw=dict(color="orchid"),
        )
    plt.savefig(
        os.path.join(
            test_directory, "demonstration-of-plot-lightcurves-and-plot-spectra.pdf"
        )
    )


def test_add_labels_to_panels():
    plt.close("all")

    fi, ax = plt.subplots(3, 3)
    _add_panel_labels(ax, preset="inside", color="blue")
    plt.savefig(
        os.path.join(
            test_directory, "demonstration-of-adding-abcde-labels-to-panels.pdf"
        )
    )


def test_pcolormesh():
    plt.close("all")

    s = (
        SimulatedRainbow(R=10, dt=10 * u.minute)
        .inject_transit()
        .inject_noise(signal_to_noise=1000)
    )
    fi, ax = plt.subplots(2, 2, constrained_layout=True, figsize=(8, 5))
    s.imshow(ax=ax[0, 0])
    plt.title("imshow")
    s.pcolormesh(ax=ax[0, 1])
    plt.title("pcolormesh")
    plt.yscale("log")
    b = s.bin(dw=0.5 * u.micron)
    b.imshow(ax=ax[1, 0])
    b.pcolormesh(ax=ax[1, 1])
    plt.savefig(
        os.path.join(test_directory, "demonstration-of-pcolormesh-vs-imshow.pdf")
    )


def test_plot_noise_comparison():
    fi, ax = plt.subplots(
        2,
        2,
        sharey="row",
        figsize=(8, 6),
        sharex=True,
        dpi=300,
        constrained_layout=True,
    )
    for i, a in enumerate([0, 0.01]):
        s = (
            SimulatedRainbow(dw=0.1 * u.micron)
            .inject_systematics(amplitude=a)
            .inject_noise()
        )
        s.imshow(ax=ax[0, i], xaxis="wavelength")
        s.plot_noise_comparison(ax=ax[1, i])
    ax[0, 0].set_title("No Systematics")
    ax[0, 1].set_title("With Systematics")
    plt.savefig(
        os.path.join(test_directory, "demonstration-of-plot-noise-comparison.pdf")
    )


def test_plot_noise_comparison_in_bins():
    plt.close("all")

    s = SimulatedRainbow(dw=0.1 * u.micron).inject_systematics().inject_noise()
    s.plot_noise_comparison_in_bins()
    plt.savefig(
        os.path.join(
            test_directory, "demonstration-of-plot-noise-comparison-in-bins.pdf"
        )
    )


def test_plot_histogram():
    plt.close("all")

    s = SimulatedRainbow(R=5).inject_noise()
    fi, ax = plt.subplots(
        s.nwave, 1, figsize=(4, 12), sharex=True, sharey=True, constrained_layout=True
    )
    for i in range(s.nwave):
        s.plot_histogram(i, ax=ax[i], expected=True)
        if i < (s.nwave - 1):
            plt.xlabel("")
    plt.savefig(os.path.join(test_directory, "demonstration-of-histogram.pdf"))



from ..rainbows import *
from .setup_tests import *
from ..rainbows.helpers.history import represent_as_copypasteable


def test_history():
    np.random.seed(0)
    x = SimulatedRainbow().inject_transit().inject_noise().bin(R=5).normalize()
    h = x.history()

    for k in ["Rainbow", "inject_transit", "bin", "normalize"]:
        assert k in h

    np.random.seed(0)
    new = eval(h)
    assert x == new


def test_history_with_slicing_and_addition():
    a = SimulatedRainbow().inject_transit().normalize()[:10:2, :]
    b = a * 2
    c = a + b
    d = eval(c.history())
    assert c == d
    assert d == a * 3


def test_represent_as_copypasteable():
    for x_in in [
        np.array([1, 2]) * u.day,
        [1, 2],
        [1, 2] * u.day,
        1 * u.day,
        np.linspace(0, 1, 10),
    ]:
        string = represent_as_copypasteable(x_in)
        x_out = eval(string)
        match = np.all(np.isclose(x_in, x_out))
        print(f"{x_in} == {x_out}? {match}")
        assert match
